from __future__ import print_function

import os, time, re, random
import tempita

class PageTemplate(dict):
    """ Namespace object for Page rendering.
    """
    # defaults for optional variables used in top level template
    keywords = ''
    search = ''
    news = ''
    # mutable class attribute for tracking handled files on a path
    _handled = {}

    def __init__(self, options, dirpath, dirnames, filenames, parent):
        # meta data
        self.template = options.template
        self.dest = options.dest
        self.options = options
        # filesystem data
        self.dirpath = dirpath
        self.dirnames = dirnames
        self.filenames = filenames
        # id/name of page
        self.basename = os.path.basename(dirpath)
        # tracks files accessed via template
        self._handled.setdefault(self.dirpath, {})
        # set parent/child relation
        self['parent'] = parent
        if isinstance(parent, dict):
            parent[self.basename] = self
        # add self to dict for template method access
        self['this'] = self

    def get(self, name, _default=KeyError):
        """ Poll various sources and return data; else KeyError.
        Checks (and immediately returns) in this order;
        - template namespace/cache
        - plugins
        - files/templates
        """
        # sub-pages, other keywords
        if self.has_key(name):
            return dict.get(self, name)
        # check for plugins
        if self.options.plugins.hasPlugin(name):
            return self.options.plugins.getPlugin(name, self)
        # sub-pages/sub-tempates
        if self.hasTemplate(name):
            return self.getTemplate(name)
        if _default is not KeyError:
            return _default
        # hack to get old attribute default values working
        if name in ['keywords', 'search', 'news']:
            return ''
        raise KeyError(name)
    __getitem__=get

    def __getattr__(self, name, default=KeyError):
        """ We want template based attribute lookup to follow the same rules.
        """
        try:
            return self.get(name, default)
        except KeyError:
            raise AttributeError(name)

    def hasTemplate(self, name):
        """ Check for matching file/template.
        """
        # sub-pages/sub-tempates
        for fname in [name, ('%s.html'%name), ('%s.tmpl'%name)]:
            if fname in self.filenames:
                return True

    def getTemplate(self, name):
        """ Get (rendered) matching file/template.
        """
        # sub-pages/sub-tempates
        for fname in [name,('%s.html'%name), ('%s.tmpl'%name)]:
            if fname in self.filenames:
                self._handled[self.dirpath][name] = None
                filepath = os.path.join(self.dirpath, fname)
                page = self._render(filepath)
                return page

    #######################################################################
    ### For generating
    #######################################################################
    def __repr__(self):
        """ Useful when debugging
        """
        return "<%s:%s:%s:%s>" % (self.__class__.__name__,
                self.template, self.dirpath, self.keys())

    def __unicode__(self):
        """ The template uses this as the method to render variables.
        """
        return self._render(self.template)

    def _render(self, template):
        tmpl = tempita.Template.from_filename(
                template, encoding='utf8')
        return tmpl.substitute(self)

    def publish(self):
        """ Render template to destination file and copy auxilary files.
        """
        #print('publish',self.dirpath)
        rootless_dirpath = '/'.join((self.dirpath.split('/')[1:]))
        savepath = os.path.join(self.dest, rootless_dirpath)
        if not os.path.exists(savepath):
            os.mkdir(savepath)
        savefile = os.path.join(savepath, 'index.html')
        # 
        if self.get('index_html', None):
            try:
                content = unicode(self).encode('ascii', 'xmlcharrefreplace')
                open(savefile, 'w').write(content)
            except:
                if os.path.exists(savefile):
                    os.remove(savefile) # remove bad savefile
                print('not rendered', savefile)
                raise
        # needs to be after respond() call
        self.copy_files(savepath) # copy over auxilary files
        self.clean_cruft(savepath) # remove unneeded files

    def clean_cruft(self, path, join=os.path.join):
        """ Remove files from dest that are no longer in source.

            Should this remove _handled files (that used to be unhandled)?
        """
        for filename in os.listdir(path):
            if filename == 'index.html':
                continue
            if filename in self.filenames or filename in self.dirnames:
                continue
            filepath = os.path.join(path, filename)
            if not os.access(filepath, os.W_OK):
                continue
            if os.path.isdir(filepath):
                for dp,dns,fns in os.walk(filepath, topdown=False):
                    for fn in fns:
                        os.remove(join(dp, fn))
                        #print('removing',join(dp, fn))
                    for dn in dns:
                        os.rmdir(join(dp, dn))
                        #print('removing',join(dp, dn))
                #print('removing', filepath)
                os.rmdir(filepath)
            else:
                #print('removing', filepath)
                os.remove(filepath)

    def copy_files(self, savepath):
        """ Copy non-template files.
        """
        def _update_if_modified(source, dest, modified = os.path.getmtime):
            if not os.path.exists(dest) or modified(source)>modified(dest):
                #print('overwriting',dest,'with',source)
                open(dest, 'w').write(open(source).read())
        basename = os.path.basename
        join = os.path.join
        for filename in self.unhandled():
            # in case you want your template under root dir
            if filename == os.path.basename(self.options.template): continue
            source = join(self.dirpath,filename)
            dest = join(savepath, filename)
            exists = os.path.exists
            islink = os.path.islink
            if islink(source): # support sym-links
                linked_source = os.readlink(source)
                local_source = join(self.dirpath, basename(linked_source))
                if exists(local_source) and not islink(local_source):
                    source = basename(linked_source)
                    if not os.path.exists(dest):
                        os.symlink(source, dest)
                else:
                    _update_if_modified(linked_source, dest)
            else:
                _update_if_modified(source, dest)

    def unhandled(self):
        """ Auxilary files not used directly in templates (eg. images).
        """
        handled = self._handled[self.dirpath]
        for filename in self.filenames:
            if not handled.has_key(filename):
                yield filename

    def base_path(self):
        """ Path for page rooted for site.
            With trailing slash (base href style).
        """
        path = self.site_path()
        trailing_slash = (path != '/') and '/' or ''
        return path + trailing_slash

    def site_path(self):
        """ Path for page rooted for site.
            Used to build paths in the templates.
        """
        if not self.dirpath.count('/'):
            return '/'
        return self.dirpath[self.dirpath.find('/'):]

    #######################################################################
    ### Used from templates 
    ######################################################################
    def title(self):
        """ Always want a title, even if not explicitally set.
        """
        if not self.has_key('title'):
            self['title'] = unicode(
                    self.get('title', self.basename.capitalize())
                    ).strip()
        return self['title']

    def breadcrumbs(self, links=1):
        """ Trail from root of site.
        """
        if not self.dirpath.count('/'): return ''
        crumbs = [self.title()]
        parent = self.get('parent', None)
        while parent is not None:
            if links:
                crumbs.append(parent.link())
            else:
                crumbs.append(parent.title())
            parent = parent.get('parent', None)
        crumbs.reverse()
        return crumbs

    def link(self):
        """ Relative link to this page.
        """
        return '<a href="%s">%s</a>' % (self.base_path(), self.title())

    def current_date(self, format="%b %d %Y"):
        return time.strftime(format,time.localtime())

    def last_modified(self, format=" %Y.%m.%d"):
        pjoin = os.path.join
        realpath = os.path.realpath
        getmtime = os.path.getmtime
        if self.template == self.options.template:
            for index in ['index_html', 'index.html', 'index.tmpl']:
                index = realpath(pjoin(self.dirpath, index))
                if os.path.exists(index): break
            else:
                # warning
                print("Missing index file", self.dirpath, index)
        else:
            index = realpath(pjoin(self.dirpath, self.template))
        files = [index]
        files.extend(pjoin(self.dirpath,f) for f in self._handled[self.dirpath])
        mtime = max(getmtime(realpath(f)) for f in files)
        return time.strftime(format, time.localtime(mtime))

    def year(self, format="%Y"):
        return time.strftime(format, time.localtime())

    def menus(self):
        """ Return all files that start with 'menu'. Use this as a
            way to get content for right menus.
        """
        menus = [fn for fn in self.filenames
                if fn.startswith('menu')]
        menus.sort()
        return [self.get(m) for m in menus]

    findnames = re.compile('<a name="(.+)">(.+)</a>', re.I).findall
    def name_links(self, findnames=findnames):
        """ #name style links
        """
        link = '<a href="#%s">%s</a>'
        return [link % (name, title)
                for (name, title) in findnames(unicode(self.get('index_html')))]

    def subdirs(self, order=None):
        """ All children/subdirectory pages.
        """
        dirnames = self.dirnames
        if order:
            dirnames = dirnames[:]
            try: # order is list in desired order
                dirnames.sort(lambda x,y: cmp(order.index(x), order.index(y)))
            except ValueError:
                print('Error ordering subdirs for: %s' % self.dirpath)
        for dirname in dirnames:
            if not self[dirname].hidden():
                yield self[dirname]

    def subdir_links(self, order=None):
        """ links for sub-directories
        """
        return [subdir.link() for subdir in self.subdirs(order=order)]

    def root(self, _cache={}):
        """ Root page.
        """
        if not _cache.has_key('__root'):
            parent = self
            while parent.get('parent', None) is not None:
                parent = parent['parent']
            _cache['__root'] = parent
        return _cache['__root']

    getquotes = re.compile('<blockquote>(.*?)(<p.*?)</blockquote>',re.S).findall
    def random_quote(self, getquotes=getquotes):
        """ I have a file full of various quotes I like. This lets me grab a
            random one for display.
        """
        quotes = self.root().get('quotes')
        quote_list = getquotes(unicode(quotes))
        assert quote_list, 'Problem parsing quotes file.'
        (quote, cite) = random.choice(quote_list)
        quote = ('<blockquote class="quote"><div>'
                    '<a href="%s">%s</a>%s</div></blockquote>'
                % (quotes.base_path(), quote, cite))
        return quote

    def fflag(self, flag, recursive=True):
        """ Flag files. If file exists, flag is true, else false.
        """
        page = self
        while page is not None:
            if flag in page.filenames:
                return True
            page = page.get('parent', None) if recursive else None
        return False


