import os, sys, inspect
import optparse, ConfigParser

from contextlib import contextmanager
from functools import wraps

from itertools import chain

def _load_options():
    """ Process and store all config and command line options.
    """
    config_options = ['root', 'dest', 'template']

    parser = optparse.OptionParser()
    # Config Options optionally kept in conf file
    parser.add_option("-r", "--root", dest="root", default="Root",
            help="Set root directory name of site to ROOT.",
            metavar="ROOT")
    parser.add_option("-d", "--dest", dest="dest",
            default="/var/www",
            help="Render site to  local directory DEST.", metavar="DEST")
    parser.add_option("-t", "--template", dest="template",
            default="layout.tmpl",
            help="Base template of site given as TMPL", metavar="TMPL")
    # Command line only arguments
    parser.add_option("-c", "--config", dest="config_file",
            default="site.config",
            help="Read config from CONFIG", metavar="CONFIG")
    parser.add_option("-g", "--generate-config", action="store_true",
            dest="generate", default=False,
            help="(Re)Generate a config file.")
    parser.add_option("-p", "--plugin-dir",
            dest="plugin_dir", default="./scripts/plugins",
            help="Directory containing plugins.")
    cmd_options = parser.parse_args()[0]
    cmd_settings = _nonDefaults(cmd_options, parser)

    # pull values from config file
    config =  ConfigParser.ConfigParser()
    config.read(cmd_options.config_file)
    if not config.has_section('site'):
        config.add_section('site')
        for opt in config_options:
            config.set('site', opt, getattr(cmd_options, opt))
    elif cmd_settings:
        for opt,val in config.items('site'):
            if opt in cmd_settings:
                config.set('site', opt, getattr(cmd_options, opt))
            else:
                setattr(cmd_options, opt, val)

    # generate new config file
    if cmd_options.generate:
        fp = open(cmd_options.config_file,'w')
        try:
            config.write(fp)
        finally:
            fp.close()

    # load and save plugins for later use
    cmd_options.plugins = _get_plugins(cmd_options.plugin_dir)

    # This needs to be run in the directory above the root of the site.
    # So lets check change directory as appropriate.
    if cmd_options.root not in os.listdir('.'):
        raise RuntimeError(('Root directory (%s) must be a subdirectory '
                            'of the current directory.') % cmd_options.root)

    return cmd_options

def _nonDefaults(options, parser):
    """Return list of settings that differ from defaults.
    Used for determining what to save in the config file.
    """
    return [option.dest for option in parser.option_list
            if option.dest and option.default != getattr(options, option.dest)]

def _get_plugins(path):
    """ Load plugins and return map of exported functions/objects.
    """
    path = os.path.abspath(path)
    with _sys_path_insert(path):
        modules = [__import__(os.path.splitext(mod)[0])
                for mod in os.listdir(path)
                if (mod.endswith('.py') or os.path.isdir(mod))
                    and not mod.endswith('plugin_types.py')]
    plugins = Plugins()
    for mod in modules:
        plugins.findPlugins(mod)
    return plugins

@contextmanager
def _sys_path_insert(path):
    """ Insert given path into sys.path w/in context.
    """
    sys.path.insert(0, path)
    yield
    sys.path.pop(0)

class Plugins(object):
    """ Class used for importing/registering plugins.

    Simply add your functions to the appropriate data structure.
    Hooks are run in list order.

    plugin          - functions inserted into template rendering namespace
                    - function parameters: page

    pre_site_hook   - functions called before any rendering has occured
                    - function parameters: page_tree, plugin_object

    pre_page_hook   - functions called before each page rendering
                    - function parameters: page, plugin_object

    post_page_hook  - functions called after each page rendering
                    - function parameters: page, plugin_object

    post_site_hook  - functions called after all page rendering has occured
                    - function parameters: page_tree, plugin_object
    """
    def __init__(self):
        """ Init list stores for different plugin types.
        """
        self.plugin         = []
        self.pre_site_hook  = []
        self.pre_page_hook  = []
        self.post_page_hook = []
        self.post_site_hook = []
        # dict to look up specific plugins
        self.namespace      = {}

    def findPlugins(self, module):
        """ Process module looking for plugins.

        Check module for functions or hooks or classes. In case of classes
        instantiate class and check instance for functions/hooks. Also treat
        class instance as possible function/call if it has __call__.
        """
        sources = [module]
        while sources:
            for (name, obj) in _get_public(sources.pop(), module):
                if type(obj) is type:
                    obj = obj()
                    sources.append(obj)
                    if hasattr(obj, 'plugin_type'):
                        if not hasattr(obj.__call__, 'plugin_type'):
                            # put on __call__.im_func like decorator would
                            obj.__call__.im_func.plugin_type = obj.plugin_type
                    # this allows for simplifying plugin wrapping code
                    obj = obj.__call__
                if not callable(obj): continue
                plugin_type = getattr(obj, 'plugin_type', 'plugin')
                #print '>', name, `obj`, plugin_type
                assert hasattr(self, plugin_type), \
                        "Bad plugin type: " + plugin_type
                getattr(self, plugin_type).append((name, obj))

    def hasPlugin(self, name):
        """ Check for plugin by name.
        """
        if not self.namespace:
            self.namespace = dict(self.plugin)
        return name in self.namespace

    def getPlugin(self, name, context):
        """ Get a function to use in a template.
        """
        if not self.namespace:
            self.namespace = dict(*self.plugin)
        func = self.namespace.get(name)
        @wraps(func)
        def _contextualize():
            return func(context)
        return _contextualize

    def getHook(self, ptype):
        """ Return hooks of given type.
        """
        assert hasattr(self, ptype), "Bad Hook type request"
        return getattr(self, ptype)

    def getHooks(self):
        """ Return all hooks.
        4-tuple of; (pre_site, pre_page, post_page, post_site)
        """
        def strip_names(hooks):
            return [h[1] for h in hooks]
        return map(strip_names,
                (self.pre_site_hook,
                self.pre_page_hook,
                self.post_page_hook,
                self.post_site_hook))

def _get_public(obj, mod):
    """ Get public functions/methods on obj (class or module).
    """
    def pred((name, obj)):
        return (not name.startswith('_') and inspect.getmodule(obj) is mod)
    return filter(pred, ((o, getattr(obj, o)) for o in dir(obj)))

# imported by other modules to get options
options = _load_options()

